import asyncio
from . import coordinator_service_pb2
from .base_extractor import ExtractorWrapper, ExtractorDescription
from typing import Optional, Tuple
from .base_extractor import Content
import nanoid
import json
from .extractor_worker import ExtractorModule, create_executor, describe
from .agent import ExtractorAgent


def local(extractor: str, text: Optional[str] = None, file: Optional[str] = None):
    if text and file:
        raise ValueError("You can only pass either text or file, not both.")
    if not text and not file:
        raise ValueError("You need to pass either text or file")
    if text:
        content = Content.from_text(text)
    if file:
        content = Content.from_file(file)
    module, cls = extractor.split(":")
    wrapper = ExtractorWrapper(module, cls)
    result = wrapper.extract_batch({"task_id": content}, params="{}")
    print(result)


def split_validate_extractor(name: str) -> Tuple[str, str]:
    try:
        module, cls = name.split(":")
    except ValueError:
        raise ValueError(
            "The extractor name should be in the format 'module_name:class_name'"
        )
    return module, cls


def join(
    extractor: str,
    workers: int,
    coordinator: str = "localhost:8950",
    ingestion_addr: str = "localhost:8900",
):
    print(f"joining {coordinator} and sending extracted content to {ingestion_addr}")
    module, cls = split_validate_extractor(extractor)
    extractor_module = ExtractorModule(module_name=module, class_name=cls)
    executor = create_executor(extractor_module, workers=workers)
    asyncio.set_event_loop(asyncio.new_event_loop())
    description: ExtractorDescription = asyncio.get_event_loop().run_until_complete(
        describe(asyncio.get_event_loop(), executor)
    )
    embedding_schemas = {}
    metadata_schemas = {}
    for name, embedding_schema in description.embedding_schemas.items():
        embedding_schemas[name] = embedding_schema.model_dump_json()
    for name, metadata_schema in description.metadata_schemas.items():
        metadata_schemas[name] = json.dumps(metadata_schema)

    print(f"embedding schemas are {embedding_schemas}")
    print(f"metadata schemas are {metadata_schemas}")

    api_extractor_description = coordinator_service_pb2.Extractor(
        name=description.name,
        description=description.description,
        input_params=description.input_params,
        input_mime_types=description.input_mime_types,
        metadata_schemas=metadata_schemas,
        embedding_schemas=embedding_schemas,
    )
    id = nanoid.generate()
    print(f"extractor id is {id}")
    server = ExtractorAgent(
        id, api_extractor_description, coordinator, executor, ingestion_addr
    )
    try:
        asyncio.get_event_loop().run_until_complete(server.run())
    except asyncio.CancelledError as ex:
        print("exiting gracefully", ex)


def describe_sync(extractor):
    module, cls = extractor.split(":")
    wrapper = ExtractorWrapper(module, cls)
    print(wrapper.describe())



from transformers import AutoTokenizer, AutoModel
import torch
import torch.nn.functional as F

from typing import List


def mean_pooling(model_output, attention_mask):
    token_embeddings = model_output[
        0
    ]  # First element of model_output contains all token embeddings
    input_mask_expanded = (
        attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
    )
    return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
        input_mask_expanded.sum(1), min=1e-9
    )


class SentenceTransformersEmbedding:
    def __init__(self, model_name) -> None:
        self._model_name = model_name
        self._tokenizer = AutoTokenizer.from_pretrained(
            f"sentence-transformers/{model_name}"
        )
        self._model = AutoModel.from_pretrained(f"sentence-transformers/{model_name}")

    def embed_ctx(self, inputs: List[str]) -> List[List[float]]:
        result = self._embed(inputs)
        return result.tolist()

    def embed_query(self, query: str) -> List[float]:
        result = self._embed([query])
        return result[0].tolist()

    def _embed(self, inputs: List[str]) -> torch.Tensor:
        encoded_input = self._tokenizer(
            inputs, padding=True, truncation=True, return_tensors="pt"
        )
        with torch.no_grad():
            model_output = self._model(**encoded_input)
        sentence_embeddings = mean_pooling(
            model_output, encoded_input["attention_mask"]
        )
        return F.normalize(sentence_embeddings, p=2, dim=1)

    def tokenizer_encode(self, inputs: List[str]) -> List[List[int]]:
        return self._tokenizer.batch_encode_plus(inputs)["input_ids"]

    def tokenizer_decode(self, tokens: List[List[int]]) -> List[str]:
        return self._tokenizer.batch_decode(tokens, skip_special_tokens=True)

    def tokenize(self, inputs: List[str]) -> List[List[str]]:
        result = []
        for input in inputs:
            result.extend(self._tokenize(input))
        return result

    def _tokenize(self, input: str) -> List[str]:
        max_length = self._tokenizer.model_max_length
        chunks = []
        chunk = []
        chunk_length = 0
        words = input.split(" ")

        for word in words:
            tokens = self._tokenizer.tokenize(word)
            token_length = len(tokens)

            if chunk_length + token_length <= max_length:
                chunk.append(word)
                chunk_length += token_length
            else:
                chunks.append(" ".join(chunk))
                chunk = [word]
                chunk_length = token_length

        chunks.append(" ".join(chunk))
        return chunks
